#ifndef COMMON_H
#define COMMON_H

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <utmp.h>

#define MAX_USERS		256
#define MAX_PROCS		(MAX_USERS * 4)
#define MAX_BOOT_WARN_TIMES	16
#define MAX_TIME_CLASSES       	4
#define FLG_PRIORITY		0xe000
#define	FLG_SMARTBOOT		0x1000
#define FLG_SSMARTBOOT		0x0800
#define FLG_IBOOT		0x0400
#define FLG_ISMARTBOOT		0x0200
#define FLG_TCSMARTBOOT		0x0100
#define FLG_EXPIRE		0x0080
#define PRIORITY(f)		(((f) & FLG_PRIORITY) >> 13)
#define SMARTBOOT(f)		((f) & FLG_SMARTBOOT)
#define	SSMARTBOOT(f)		((f) & FLG_SSMARTBOOT)
#define IBOOT(f)		((f) & FLG_IBOOT)
#define ISMARTBOOT(f)		((f) & FLG_ISMARTBOOT)
#define TCSMARTBOOT(f)		((f) & FLG_TCSMARTBOOT)
#define EXPIRE(f)		((f) & FLG_EXPIRE)
#define EXPIRE_DELETE		0x0000
#define EXPIRE_UNSUBSCRIBE	0x0080
#define min(x,y)		((x) < (y) ? (x) : (y))
#define max(x,y)		((x) > (y) ? (x) : (y))

typedef unsigned char		byte;
typedef unsigned int		word;
typedef unsigned long long	dword;

typedef struct {
  uid_t		uid				__attribute__ ((packed));
  ushort       	flags				__attribute__ ((packed));
  word		phNoArea			__attribute__ ((packed));
  word		phNoLocal			__attribute__ ((packed));
  int		tLeft				__attribute__ ((packed));
  int		tLimit				__attribute__ ((packed));
  int		sLeft				__attribute__ ((packed));
  int		sLimit				__attribute__ ((packed));
  short		cLeft[MAX_TIME_CLASSES]		__attribute__ ((packed));
  short		cLimit[MAX_TIME_CLASSES]	__attribute__ ((packed));
  time_t	subExpire			__attribute__ ((packed));
  ushort       	subFlags			__attribute__ ((packed));
  int		subTlimit			__attribute__ ((packed));
  int		subSlimit			__attribute__ ((packed));
  short		subClimit[MAX_TIME_CLASSES]	__attribute__ ((packed));
  dword		sBytes				__attribute__ ((packed));
  dword		tBytes				__attribute__ ((packed));
  time_t	expire				__attribute__ ((packed));
  time_t	last				__attribute__ ((packed));
} UserRec;	/* sizeof(UserRec) = 90 */

typedef struct {
  byte		flags		__attribute__ ((packed));
  word		phNoArea       	__attribute__ ((packed));
  word		phNoLocal      	__attribute__ ((packed));
  time_t	expire		__attribute__ ((packed));
} BanRec;

typedef struct {
  pid_t		pid;
  uid_t		uid, euid;
  ushort	tty;
  char		name[16];
} ProcRec;

typedef struct {
  uid_t		uid;
  char		tty[16];
  time_t	time;
} LoginRec;

typedef struct {
  int		startDay, endDay;
  int		startHour, endHour;
} TimeClass;

void		errQuit(char *format, ...);
void		perrQuit(char *name);
int		runCommand(char *path, ...);
time_t		addTime(char *s);
void		procList(int *nProcs, ProcRec *procRec);
int		allLinesBusy(int nProcs, ProcRec *procRec, int devBase, int nLines);
void		userList(int *nLogins, LoginRec *loginRec);
int		userLoggedIn(int nLogins, LoginRec *loginRec, uid_t uid);
int		getToken(char *d, char *s, int *idx);
void		fCopy(char *spath, char *dpath);
uid_t		UIDfromLogin(char *login);
char		*loginFromUID(char *login, uid_t uid);
void		userFileOpen();
void		userFileClose();
int		userFileRead(UserRec *ur);
int		userFileWrite(UserRec *ur);
void		userFileEdit(UserRec *ur);
int		userFileSearch(UserRec *ur, uid_t uid);
void		banFileOpen();
void		banFileClose();
int		banFileRead(BanRec *br);
int		banFileWrite(BanRec *br);
int		banFileSearch(BanRec *br, word phNoArea, word phNoLocal);
int		phNoToWords(word *phNoArea, word *phNoLocal, char *phNo, int phNoDigits);
char		*wordsToPhNo(char *phNo, word phNoArea, word phNoLocal, char *phNoAreaFormat, char *phNoLocalFormat);
int		stringSum(char *s);
void		bytes2ASCII(char *str, dword bytes);
int		curTimeClass(int nTimeClasses, TimeClass *timeClass);
void		processConfig();

extern const char	*configFilePath;
extern const char	*userFilePath;
extern FILE		*userFile;
extern const char	*banFilePath;
extern FILE		*banFile;
extern int		multiBase, nLines;
extern int		optReturnDelay;
extern int		optSendMail;
extern int		optGuestTime;
extern int		optGuestPriority;
extern int		nBootWarnTimes;
extern int		bootWarnTime[MAX_BOOT_WARN_TIMES];
extern int		optBootOne;
extern int		optSmartBoot;
extern int		optSessionSmartBoot;
extern int		optIdleBoot;
extern int		optIdleSmartBoot;
extern int		nTimeClasses;
extern TimeClass	timeClass[MAX_TIME_CLASSES];
extern int		optTimeClassSmartBoot;
extern int		nExcluded;
extern uid_t		excluded[64];
extern char		modemDial[32];
extern char		phNoAreaFormat[16];
extern char		phNoLocalFormat[16];
extern int		phNoDigits;

#endif
