#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <limits.h>
#ifdef LINUX
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/ppp_defs.h>
#endif
#ifdef BSD
#include <net/if.h>  
#include <net/if_ppp.h>  
#include <net/ppp_defs.h>  
#endif
#include <paths.h>
#include <pwd.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/dir.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <syslog.h>
#include <utime.h>
#include "common.h"

#define		MAX_PPP_UNITS			MAX_USERS
#define		MSG_BUSY			0
#define		MSG_IDLE			1
#define		MSG_TIME_CLASS			2
#define		MSG_TIME_PERIOD			3
#define		MSG_TIME_SESSION       		4

void		daemonInit();
void		update();
char		*userTTY(char *devPath, int uid);
void		userMessage(char *login, char *devName, char *format, ...);
int		devNumFromName(char *devName);
int		PPPidleLeft(char *ttyPath);
void		PPPupdate(UserRec *ur, int unit);
void		PPPfindUnits();
int		PPPfindUnit(uid_t uid);
void		logQuit(char *format, ...);

const char	*msgTimeWarning = "You have %d minute(s) remaining.";
const char	*msgIdleWarning = "You will be logged out after 1 more minute of idle time.";
const char	*msgPath[5] = { LIB"/user_updated.busy", LIB"/user_updated.idle", LIB"/user_updated.time_class", LIB"/user_updated.time_period", LIB"/user_updated.time_session" };

char		MAILHOST[64];
time_t		configTime;
int		allBusy;
int		nLogins;
LoginRec	loginRec[MAX_USERS];
int		nProcs;
ProcRec		procRec[MAX_PROCS];
uid_t		guestUID;
pid_t		PPPuser[MAX_PPP_UNITS];
int		nKill;
pid_t		killPID[MAX_PROCS];

void main() {
  struct stat _stat;
  time_t t;
  struct tm *tmP;
  processConfig(); configTime = time(NULL);
  daemonInit();
  openlog("user_updated", 0, LOG_DAEMON);
  guestUID = UIDfromLogin("guest");
  if (getenv("MAILHOST")) strcpy(MAILHOST, getenv("MAILHOST"));
  for (;;) {
    if (stat(configFilePath, &_stat) < 0)
      logQuit("couldn't stat %s", configFilePath);
    if (_stat.st_mtime > configTime) {
      processConfig();
      if (configTime)
        syslog(LOG_INFO, "re-read %s", configFilePath);
      configTime = _stat.st_mtime;
    }
    userList(&nLogins, loginRec);
    procList(&nProcs, procRec);
    PPPfindUnits();
    allBusy = allLinesBusy(nProcs, procRec, multiBase, nLines);
    update();
    if (nKill) {
      int i;
      sleep(5);
      for (i = 0; i < nKill; i++)
	kill(killPID[i], SIGKILL);
    }
    t = time(NULL);
    tmP = localtime(&t);
    sleep(60 - tmP->tm_sec);
  }
}

void daemonInit() {
  if (fork()) exit(0);
  setsid();
  chdir("/");
  umask(0);
}

void update() {
  char login[UT_NAMESIZE + 1], devPath[16], str[256];
  int minPriority = 7, minSleft = INT_MAX;
  int i, j, k, PPPunit, idleLeft, sLeft, guestOnline;
  int curClass = curTimeClass(nTimeClasses, timeClass);
  struct stat _stat;
  UserRec ur;
  int nVictims = 0;
  uid_t victim[MAX_USERS];
  byte victimMsg[MAX_USERS];
  nKill = 0;
  for (guestOnline = 0, i = 0; i < nProcs; i++)
    if (procRec[i].uid == guestUID) guestOnline = 1;
  userFileOpen();
  while (!userFileRead(&ur)) {
    for (i = 0; i < nProcs && ur.uid != procRec[i].uid && ur.uid != procRec[i].euid; i++);
    if (i == nProcs) continue;
    loginFromUID(login, ur.uid);
    if (!userTTY(devPath, procRec[i].uid)) continue;
    PPPunit = PPPfindUnit(ur.uid);
    /* update their record */
    if (curClass < 0 || ur.cLimit[curClass] > -3)
      ur.tLeft--;
    if (curClass < 0 || ur.cLimit[curClass] > -2)
      ur.sLeft--;
    if (curClass >= 0 && ur.cLimit[curClass] >= 0)
      ur.cLeft[curClass]--;
    time(&ur.last);
    if (PPPunit >= 0) PPPupdate(&ur, PPPunit);
    userFileEdit(&ur);
    sLeft = INT_MAX;
    /* figure out how much time they have left */
    if ((!SMARTBOOT(ur.flags) || allBusy) && ur.tLimit >= 0) sLeft = ur.tLeft;
    if ((!SSMARTBOOT(ur.flags) || allBusy) && ur.sLimit >= 0) sLeft = min(sLeft, ur.sLeft);
    if ((!TCSMARTBOOT(ur.flags) || allBusy) && curClass >= 0 && ur.cLimit[curClass] >= 0)
      sLeft = min(sLeft, ur.cLeft[curClass]);
    if (!PRIORITY(ur.flags) && allBusy) sLeft = min(sLeft, 0);
    /* figure out how much idle time they have left */
    if (optIdleBoot && IBOOT(ur.flags)) {
      if (PPPunit >= 0) idleLeft = PPPidleLeft(devPath);
      else {
	if (stat(devPath, &_stat)) continue;
	idleLeft = optIdleBoot * 60 - (time(NULL) - _stat.st_atime);
      }
      if (idleLeft < 0) idleLeft = 0;
      else idleLeft = idleLeft / 60 + 1;
    } else idleLeft = INT_MAX;
    /* send a warning message if necessary */
    if (PPPunit < 0) {
      if (idleLeft == 1 && sLeft > 1) 
	userMessage(login, devPath + 5, (char*)msgIdleWarning);
      else for (j = 0; j < nBootWarnTimes; j++)
	if (sLeft == bootWarnTime[j]) {
	  userMessage(login, devPath + 5, (char*)msgTimeWarning, sLeft);
	  break;
	}
    }
    /* take idle time into account */
    if (!ISMARTBOOT(ur.flags) || allBusy) sLeft = min(sLeft, idleLeft);
    if (sLeft <= 0) {
      if (optBootOne) {
	if (PRIORITY(ur.flags) > minPriority || (PRIORITY(ur.flags) == minPriority && sLeft > minSleft))
	  continue;
	
	minPriority = PRIORITY(ur.flags);
	minSleft = sLeft;
	nKill = 0;
	nVictims = 0;
      }
      if (optSendMail && PPPunit >= 0 && ur.tLimit > 0) {
	victim[nVictims] = ur.uid;
	if (ur.tLeft <= 0) victimMsg[nVictims++] = MSG_TIME_PERIOD;
	else if (ur.sLeft <= 0) victimMsg[nVictims++] = MSG_TIME_SESSION;
	else if (curClass >= 0 && ur.cLimit[curClass] >= 0 && ur.cLeft[curClass] <= 0) victimMsg[nVictims++] = MSG_TIME_CLASS;
	else victimMsg[nVictims++] = MSG_IDLE;
      }
    } else continue;
    for (; i < nProcs; i++)
      if ((procRec[i].uid == ur.uid) || (procRec[i].euid == ur.uid))
        killPID[nKill++] = procRec[i].pid;
  }
  userFileClose();
  for (i = 0; i < nLogins; i++) {
    if (loginRec[i].uid != guestUID) continue;
    sLeft = (optGuestTime * 60 - (time(NULL) + 5 - loginRec[i].time)) / 60 + 1;
    if (!optGuestPriority && allBusy) sLeft = min(sLeft, 0);
    if (optIdleBoot) {
      if (stat(loginRec[i].tty, &_stat)) continue;
      idleLeft = optIdleBoot * 60 - (time(NULL) - _stat.st_atime);
      if (idleLeft < 0) idleLeft = 0;
      else idleLeft = idleLeft / 60 + 1;
    } else idleLeft = INT_MAX;
    if (idleLeft == 1 && sLeft > 1)
      userMessage("guest", loginRec[i].tty, (char*)msgIdleWarning);	
    else for (j = 0; j < nBootWarnTimes; j++)
      if (sLeft == bootWarnTime[j]) {
	userMessage("guest", loginRec[i].tty, (char*)msgTimeWarning, sLeft);
	break;
      }
    /* take idle time into account */
    if (!optIdleSmartBoot || allBusy) sLeft = min(sLeft, idleLeft);
    if (sLeft <= 0) {
      if (optBootOne) {
	if (optGuestPriority > minPriority || (optGuestPriority == minPriority && sLeft > minSleft))
	  continue;
	
	minPriority = optGuestPriority;
	minSleft = sLeft;
	nKill = 0;
	nVictims = 0;
      }
    } else continue;
    k = devNumFromName(loginRec[i].tty);
    for (j = 0; j < nProcs; j++)
      if (procRec[j].tty == k) killPID[nKill++] = procRec[j].pid;
  }
  for (i = 0; i < nKill; i++) kill(killPID[i], SIGTERM);
  for (i = 0; i < nVictims; i++) {
    char maildest[64];
    loginFromUID(maildest, victim[i]);
    if (MAILHOST[0]) sprintf(maildest + strlen(maildest), "@%s", MAILHOST);
    sprintf(str, "mail -s \"Connection Terminated\" %s < %s", maildest, msgPath[victimMsg[i]]);
    system(str);
  }
}

char *userTTY(char *devPath, int uid) {
  int i;
  for (i = 0; i < nLogins; i++)
    if (loginRec[i].uid == uid) {
      sprintf(devPath, "/dev/%s", loginRec[i].tty);
      return devPath;
    }
  devPath[0] = '\0';
  return NULL;
}

void userMessage(char *login, char *devName, char *format, ...) {
  int fd;
  FILE *f;
  char devPath[16], line[256];
  struct stat _stat;
  struct utimbuf _utime;
  va_list ap;
  va_start(ap, format);
  sprintf(devPath, "/dev/%s", devName);
  if (stat(devPath, &_stat)) return;
  if (!(_stat.st_mode & S_IWGRP)) return;
  vsprintf(line, format, ap);
  fd = open(devPath, O_WRONLY | O_NONBLOCK);
  if (fd >= 0) {
    f = fdopen(fd, "w");
    fputc(7, f);	/* send ^G */
    fprintf(f, "\n\nNOTICE: ");
    fprintf(f, "%s\n\n", line);
    fclose(f);
  }
  if (optIdleBoot) {
    _utime.actime = _utime.modtime = _stat.st_atime;
    utime(devPath, &_utime);
  }
  va_end(ap);
}

int devNumFromName(char *devName) {
  struct stat _stat;
  char devPath[16];
  sprintf(devPath, "/dev/%s", devName);
  if (stat(devPath, &_stat) < 0)
    perrQuit("stat");
  return _stat.st_rdev;
}

int PPPidleLeft(char *ttyPath) {
  struct ppp_idle _ppp_idle;
  int fd = open(ttyPath, O_RDONLY);
  if (fd < 0) return 0;
  if (ioctl(fd, PPPIOCGIDLE, &_ppp_idle) < 0) goto err;
  close(fd);
  return optIdleBoot * 60 - min(_ppp_idle.xmit_idle, _ppp_idle.recv_idle);
err:  
  close(fd);
  return 0;
}

void PPPupdate(UserRec *ur, int PPPunit) {
  struct ifreq _ifreq;
  struct ppp_stats _ppp_stats;
  int s;
  word sBytes;
  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) logQuit("couldn't create socket");
  bzero(&_ifreq, sizeof(struct ifreq));
#ifdef LINUX
  sprintf(_ifreq.ifr_ifrn.ifrn_name, "ppp%d", PPPunit);
#else ifdef BSD
  sprintf(_ifreq.ifr_name, "ppp%d", PPPunit);
#endif
  _ifreq.ifr_ifru.ifru_data = (caddr_t)&_ppp_stats;
  if (ioctl(s, SIOCGPPPSTATS, (caddr_t)&_ifreq) < 0) goto err;
#ifdef LINUX
  sBytes = _ppp_stats.p.ppp_ioctects + _ppp_stats.p.ppp_ooctects;
#else ifdef BSD
  sBytes = _ppp_stats.p.ppp_ibytes + _ppp_stats.p.ppp_obytes;
#endif
  if (sBytes > ur->sBytes) {
    ur->tBytes += sBytes - ur->sBytes;
    ur->sBytes = sBytes;
  }
err:
  close(s);
}

void PPPfindUnits() {
  int i, len, PPPunit;
  DIR *d;
  FILE *f;
  pid_t pid;
  struct dirent *de;
  int unitStrLen;
  char unitStr[8], str[32];
  for (i = 0; i < MAX_PPP_UNITS; i++) PPPuser[i] = (uid_t)-1;
  if ((d = opendir(_PATH_VARRUN)) == NULL) logQuit("opendir failed");
  while ((de = readdir(d)) != NULL) {
    len = strlen(de->d_name);
    if (len < 8 || len > 10) continue;
    if (strncmp(de->d_name, "ppp", 3)) continue;
    if (strcmp(de->d_name + len - 3, "pid")) continue;
    for (unitStrLen = 0, i = 3; isdigit(de->d_name[i]); i++)
      unitStr[unitStrLen++] = de->d_name[i];
    unitStr[unitStrLen] = '\0';
    PPPunit = atoi(unitStr);
    if (PPPunit >= MAX_PPP_UNITS) continue;
    sprintf(str, "%s/ppp%d.pid", _PATH_VARRUN, PPPunit);
    if ((f = fopen(str, "r")) == NULL) continue;
    fgets(str, 32, f);
    pid = atoi(str);
    fclose(f);
    for (i = 0; i < nProcs; i++)
      if (!strcmp(procRec[i].name, "pppd") && procRec[i].pid == pid) {
	PPPuser[PPPunit] = procRec[i].uid;
	break;
      }
  }
  closedir(d);
}

int PPPfindUnit(uid_t uid) {
  int i;
  for (i = 0; i < MAX_PPP_UNITS; i++)
    if (PPPuser[i] == uid) return i;
  return -1;
}

void logQuit(char *format, ...) {
  char buf[256];
  va_list ap;
  va_start(ap, format);
  vsprintf(buf, format, ap);
  syslog(LOG_ERR, buf);
  exit(1);
}
